# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol


class EmailSender(Protocol):
    __module__: str = 'cbra.types'

    async def send(
        self,
        sender: str,
        recipients: list[str],
        subject: str,
        content: list[str],
        headers: dict[str, Any] | None = None
    ) -> None:
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import hmac


class WebhookSecret:
    __module__: str = 'cbra.ext.picqer'

    def __init__(self, secret: bytes | str):
        if isinstance(secret, str):
            secret = str.encode(secret, 'ascii')
        assert isinstance(secret, bytes)
        self.__secret = secret

    async def verify(
        self,
        signature: bytes,
        message: bytes,
    ) -> bool:
        return hmac.compare_digest(
            hmac.digest(self.__secret, message, 'sha256'),
            signature
        )


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic


class Operation(pydantic.BaseModel):
    """This resource represents a long-running operation
    that is the result of a network API call.
    """
    name: str = pydantic.Field(
        default=...,
        title='Name',
        description=(
            'The server-assigned name, which is only unique within '
            'the same service that originally returns it.'
        )
    )

    done: bool = pydantic.Field(
        default=...,
        title='Done',
        description=(
            'If the value is `false`, it means the operation is still '
            'in progress. If `true`, the operation is completed, and '
            'either `error` or `response` is available.\n\n'
            'The value of this field is always `false` if the **Operation** '
            'object was included in a HTTP response with a 200 status code.'
        )
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools

from cbra.types import IDependant
from .models import AuthorizationServerModel
from .models import AuthorizationState


class BaseStorage(IDependant):
    __module__: str = 'cbra.ext.oauth2'

    async def destroy(self, obj: AuthorizationServerModel) -> None:
        raise NotImplementedError

    async def get_state(self, key: str) -> AuthorizationState | None:
        raise NotImplementedError

    @functools.singledispatchmethod
    async def persist(
        self,
        obj: AuthorizationServerModel
    ) -> None:
        raise NotImplementedError
    
    async def persist_state(self, obj: AuthorizationState) -> None:
        raise NotImplementedError

    @persist.register
    async def _persist_state(
        self,
        obj: AuthorizationState
    ) -> None:
        return await self.persist_state(obj)


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import copy
import types
from typing import Any
from typing import Literal
from typing import Union
from typing import get_args
from typing import get_origin

import pydantic
import pydantic.main
import pydantic.schema

from .resourceidentifier import ResourceIdentifier
from .resourcelist import ResourceList


class ResourceModelType(pydantic.main.ModelMetaclass):

    def __new__( # type: ignore
        cls,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        **params: Any
    ):
        annotations: dict[str, type] = namespace.get('__annotations__') or {}
        is_abstract = namespace.pop('__abstract__', False)
        if not is_abstract:
            # Collect readonly fields to create the CreateResourceRequest
            # ReplaceResourceRequest and UpdateResourceRequest models.
            create_fields: list[tuple[str, type, pydantic.main.FieldInfo | None]] = []
            key_fields: list[tuple[str, type, pydantic.main.FieldInfo]] = []
            update_fields: list[tuple[str, type, pydantic.main.FieldInfo | None]] = []
            response_fields: list[tuple[str, type, pydantic.main.FieldInfo | None]] = []
            for attname, class_ in annotations.items():
                field = namespace.get(attname)
                if isinstance(field, pydantic.main.FieldInfo):
                    response_fields.append((attname, class_, cls.get_field(field, 'response')))
                    if field.extra.get('primary_key'):
                        key_fields.append((attname, class_, field))
                    continue
                create_fields.append((attname, class_, cls.get_field(field, 'create')))
                update_fields.append((attname, class_, cls.get_field(field, 'update')))
                response_fields.append((attname, class_, cls.get_field(field, 'response')))

            PartialModel = type('Partial{name}Request', (pydantic.BaseModel,), {
                '__annotations__': {attname: class_ for attname, class_, _ in create_fields},
                **{name: field for name, _, field in create_fields if field is not None},
            })
            UpdateResourceRequest = type(f'Update{name}Request', (pydantic.BaseModel,), {
                '__annotations__': {attname: cls.allow_empty(class_) for attname, class_, _ in update_fields},
                **{name: field for name, _, field in update_fields if field is not None},
            })
            ResponseModel = type(name, (pydantic.BaseModel,), {
                '__annotations__': {attname: cls.ensure_required(field, class_) for attname, class_, field in response_fields},
                **{name: field for name, _, field in response_fields if name in namespace},
            })
            namespace.update({
                '__create_model__': type(f'Create{name}Request', (PartialModel,), {}),
                '__update_model__': UpdateResourceRequest,
                '__replace_model__': type(f'Replace{name}Request', (PartialModel,), {}),
                '__response_model__': ResponseModel,
                '__list_model__': type(f'{name}List', (ResourceList,), {
                    '__annotations__': {
                        'kind': Literal[f'{name}List'], # type: ignore
                        'items': list[ResponseModel]
                    },
                    'items': pydantic.Field(
                        default=...,
                        title=f'{name} array',
                        description=f'Items in the list of **{name}** resources.'
                    ),
                    'kind': pydantic.Field(
                        default=...,
                        required=[name]
                    )
                })
            })
            if key_fields:
                namespace.update({
                    '__key_model__': type(f'{name}Identifier', (ResourceIdentifier,), {
                        '__annotations__': {attname: cls.ensure_required(field, class_) for attname, class_, field in key_fields},
                        **{name: field for name, _, field in key_fields if name in namespace},
                    })
                })

        return super().__new__(cls, name, bases, namespace, **params) # type: ignore

    @staticmethod
    def allow_empty(value: Any) -> Any:
        return value | None

    @staticmethod
    def ensure_required(field: pydantic.fields.FieldInfo | None, value: Any) -> Any:
        origin = get_origin(value)
        if origin == types.UnionType and field and field.extra.get('read_only'):
            args: list[Any] = [x for x in get_args(value) if x != types.NoneType]
            if not args:
                raise TypeError("Field can not only contain None.")
            value = args[0]
            if len(args) > 1:
                value = Union[tuple(args)] # type: ignore
        return value

    @staticmethod
    def get_field(
        field: pydantic.main.FieldInfo | None,
        action: str
    ) -> pydantic.main.FieldInfo | None:
        field = copy.deepcopy(field)
        if isinstance(field, pydantic.main.FieldInfo) and field.extra.get('read_only'):
            if action not in {'response', 'key'}:
                return
            field.update_from_config({
                'default': ...,
            })
        elif isinstance(field, pydantic.main.FieldInfo) and action == 'update':
            field.update_from_config({'default': None})
        return field




# Copyright (C) 2020-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os


__all__: list[str] = [
    'GOOGLE_DATASTORE_NAMESPACE',
    'GOOGLE_HOST_PROJECT',
    'GOOGLE_SERVICE_PROJECT',
]


GOOGLE_DATASTORE_NAMESPACE: str | None = os.environ.get('GOOGLE_DATASTORE_NAMESPACE')
GOOGLE_HOST_PROJECT: str | None = os.environ.get('GOOGLE_HOST_PROJECT')
GOOGLE_SERVICE_PROJECT: str | None = os.environ.get('GOOGLE_SERVICE_PROJECT')


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from cbra.types import Abortable


class ResponseModeNotSupported(Abortable):
    status_code: int = 400

    def __init__(self, mode: str):
        self.mode = mode

    async def as_response(self) -> fastapi.Response:
        return fastapi.responses.PlainTextResponse(
            content=f'Response mode {self.mode} is not supported by this server.'
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import datetime
from typing import TypeAlias
from typing import Union

import pydantic

from cbra.types import PersistedModel
from ..types import PrincipalType

from ..principalhasher import PrincipalHasher
from .emailprincipal import EmailPrincipal
from .externalprincipal import ExternalPrincipal
from .publicidentifierprincipal import PublicIdentifierPrincipal


PrincipalSpecType: TypeAlias = Union[
    EmailPrincipal,
    ExternalPrincipal,
    PublicIdentifierPrincipal
]

hasher: PrincipalHasher = PrincipalHasher()


class Principal(PersistedModel):
    id: int | None = pydantic.Field(
        default=None,
        auto_assign=True
    )
    spec: PrincipalSpecType
    subject: int
    asserted: datetime
    suspended: bool = False

    #: A hash value created from the spec. For internal use only.
    key: str = pydantic.Field(..., primary_key=True)

    #: Indicates if the :class:`Principal` is trusted, meaning that we trust
    #: it is correct and is never overwritten with an *untrusted* principal
    #: of the same type.
    trust: bool = False

    @classmethod
    def new(
        cls,
        subject: int,
        issuer: str,
        principal: PrincipalType,
        asserted: datetime,
        trust: bool = False
    ):
        return cls.parse_obj({
            'asserted': asserted,
            'subject': subject,
            'trust': trust,
            'key': hasher.hash(principal),
            'spec': {
                'iss': issuer,
                'kind': type(principal).__name__,
                'principal': principal,
            }
        })


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""

.. _guides-picqer:

===================================
Building an application with Picqer
===================================


Settings
========

.. setting:: PICQER_WEBHOOK_SECRET

    Default: `None`

    The secret used to verify incoming webhooks from Picqer. This is the
    default secret used by :class:`~cbra.ext.picqer.PicqerWebhookEndpoint`.
"""
from cbra.core.conf import settings

from .picqerwebhookendpoint import PicqerWebhookEndpoint
from .webhooksecret import WebhookSecret
from . import v1


settings.register('PICQER_WEBHOOK_SECRET', None)


__all__: list[str] = [
    'v1',
    'PicqerWebhookEndpoint',
    'WebhookSecret',
]


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi
import logging

from cbra.core.conf import settings
from cbra.types import IDeferred
from cbra.types import IDependant
from cbra.types import ISessionFactory
from cbra.types import ISessionManager
from cbra.types import Session
from ..params import ApplicationSecretKey
from ..secretkey import SecretKey


class RequestSession(IDeferred, IDependant, ISessionFactory[Session], ISessionManager[Session]):
    __module__: str = 'cbra.core.session'
    cookie_name: str
    data: Session
    key: SecretKey
    logger: logging.Logger = logging.getLogger('uvicorn')
    request: fastapi.Request

    def __init__(
        self,
        request: fastapi.Request,
        key: SecretKey = ApplicationSecretKey
    ) -> None:
        self.cookie_name = settings.SESSION_COOKIE_NAME
        self.key = key
        self.request = request
        self.data = IDeferred.defer(self, 'data')

    async def add_to_response(self, response: fastapi.Response) -> None:
        await self.data.sign(self.key.sign)
        if settings.SESSION_COOKIE_SAMESITE is False:
            samesite = None
        else:
            samesite = settings.SESSION_COOKIE_SAMESITE
        assert samesite is None or isinstance(samesite, str) # nosec
        response.set_cookie(
            key=settings.SESSION_COOKIE_NAME,
            value=self.data.as_cookie(),
            domain=settings.SESSION_COOKIE_DOMAIN,
            max_age=settings.SESSION_COOKIE_AGE,
            httponly=settings.SESSION_COOKIE_HTTPONLY,
            secure=settings.SESSION_COOKIE_SECURE,
            path=settings.SESSION_COOKIE_PATH,
            samesite=samesite # type: ignore
        )

    async def clear(self) -> None:
        self.data = await self.create()

    async def create(self) -> Session:
        self.dirty = True
        return Session.new()

    async def initialize(self) -> None:
        if settings.SESSION_COOKIE_NAME not in self.request.cookies:
            self.data = await self.create()
        else:
            data = Session.parse_cookie(
                self.request.cookies[settings.SESSION_COOKIE_NAME]
            )
            if data is None:
                self.logger.critical(
                    'Session cookie was present but could not be deserialized.'
                )
                self.data = await self.create()
                await self.data.sign(self.key.sign)
            else:
                self.data = data
                self.logger.debug(
                    'Request included a session (id: %s, hmac: %s)',
                    self.data.id, self.data.hmac
                )
            if not await self.data.verify(self.key.verify):
                self.logger.critical(
                    'Session cookie was present but the signature did not '
                    'validate.'
                )
                self.data = await self.create()


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import collections
import itertools
import types
from typing import Any

from cbra.types import IEndpoint
from cbra.types import IRequestPrincipal
from cbra.types import RequestPrincipal
from .requesthandler import RequestHandler
from .optionsrequesthandler import OptionsRequestHandler


class EndpointType(type):
    http_methods: list[str] = [
        'options',
        'head',
        'get',
        'post',
        'patch',
        'put',
        'delete',
        'trace'
    ]

    def __new__(
        cls,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        **params: Any
    ) -> type[IEndpoint]:
        annotations = namespace.get('__annotations__') or {}
        is_abstract = namespace.pop('__abstract__', False)
        handlers: dict[str, RequestHandler[Any]] = collections.OrderedDict()
        if not is_abstract:
            for h in itertools.chain(*[getattr(x, 'handlers', {}) for x in bases]):
                handlers[h.attname] = h.clone() # type: ignore

            for attname, func in list(namespace.items()):
                if attname not in cls.http_methods:
                    continue
                if not callable(func):
                    raise TypeError("A request handler must be callable.")
                handlers[attname] = RequestHandler(
                    name=name,
                    method=attname,
                    func=func
                )

            namespace.update({
                'allowed_http_methods': [x.method for x in handlers.values()],
                'handlers': handlers.values()
            })

        # Create a concrete Principal subclass. This is to allow
        # thing like principal: RFC9068RequestPrincipal | OIDCRequestPrincipal.
        if annotations.get('principal'):
            EndpointPrincipal: type[RequestPrincipal]
            principal: types.UnionType | IRequestPrincipal = annotations['principal']
            if isinstance(principal, types.UnionType):
                EndpointPrincipal = annotations['principal'] = type(
                    f'{name}Principal',
                    (RequestPrincipal,),
                    {'__annotations__': {'__root__': principal}}
                )
                namespace['principal'] = EndpointPrincipal.depends()


        Endpoint = super().__new__(cls, name, bases, namespace, **params)
        if not is_abstract:
            if 'OPTIONS' not in namespace['allowed_http_methods']\
            and getattr(Endpoint, 'with_options', True):
                # Create an options handler using the default CORS
                # policy.
                handlers['options'] = OptionsRequestHandler(name)
                namespace['allowed_http_methods'].append('OPTIONS')
        for handler in handlers.values():
            handler.add_to_class(Endpoint) # type: ignore
        return Endpoint # type: ignore


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import itertools
import re
from typing import Any

import inflect

from .createaction import CreateAction
from .customaction import CustomAction
from .deleteaction import DeleteAction
from .listaction import ListAction
from .replaceaction import ReplaceAction
from .resourceaction import ResourceAction
from .retrieveaction import RetrieveAction
from .updateaction import UpdateAction


engine: inflect.engine = inflect.engine()


class ResourceType(type):
    __module__: str = 'cbra.core'
    actions: dict[str, type[ResourceAction]] = {
        'create'    : CreateAction,
        'destroy'   : DeleteAction,
        'list'      : ListAction,
        'replace'   : ReplaceAction,
        'retrieve'  : RetrieveAction,
        'update'    : UpdateAction,
        #'notify',
    }

    def __new__(
        cls,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        **params: Any
    ):
        actions: list[ResourceAction] = []
        is_abstract = namespace.pop('__abstract__', False)
        if not is_abstract:
            Model = params.get('model', None)
            if Model is None:
                raise TypeError('The `model` class argument is required.')

            # Collect actions and convert them to ResourceAction instances, if
            # necessary. First inspect the bases for known actions, and then
            # the namespace of the class we're creating.
            parent_actions: dict[str, ResourceAction] = {}
            collection_actions: set[str] = set()
            detail_actions: set[str] = set()

            for action_name, base in itertools.product(cls.actions, bases):
                handler = getattr(base, action_name, None)
                if handler is None:
                    continue
                parent_actions[action_name] = handler

            for action_name in cls.actions:
                action = namespace.get(action_name) or parent_actions.get(action_name)
                if action is None:
                    continue
                if not isinstance(action, ResourceAction):
                    action = cls.actions[action_name].fromfunc(name, action)
                if action.is_detail(): detail_actions.add(action.method)
                if not action.is_detail(): collection_actions.add(action.method)
                actions.append(action)

            # Get local non-default actions
            for func in namespace.values():
                if not hasattr(func, 'action'):
                    continue
                action = CustomAction.fromfunc(name, func)
                if action.is_detail(): detail_actions.add(action.method)
                if not action.is_detail(): collection_actions.add(action.method)
                actions.append(action)

            # Update the namespace with the actions and resource name.
            namespace.update({
                '__actions__': actions,
                'resource_name': Model.__name__,
            })
            
            # Some naming stuff.
            verbose_name = namespace.get('verbose_name') or namespace['resource_name']
            verbose_name_plural = namespace.get('verbose_name_plural')\
                or engine.plural_noun(verbose_name)
            namespace.update({
                'verbose_name': verbose_name,
                'verbose_name_plural': verbose_name_plural
            })

            # Construct the path name, if not provided, using the 
            # resource name.
            path_name: Any = namespace.setdefault(
                'path_name',
                engine.plural_noun(str.lower(Model.__name__))
            )
            if not isinstance(path_name, str)\
            or not re.match('[a-zA-Z0-9]', path_name[0]):
                raise TypeError(
                    f'{name}.path_name must be a string consisting of '
                    'alphanumeric characters.'
                )

        Resource = super().__new__(cls, name, bases, namespace, **params)
        if not is_abstract and actions:
            for action in actions:
                action.add_to_class(Resource) # type: ignore
        return Resource


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any


class IRequestPrincipalIntrospecter:
    """Knows how to obtain details about a specific :class:`IRequestPrincipal`
    implementation.
    """
    __module__: str

    async def introspect(self, principal: Any) -> Any:
        raise NotImplementedError


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import fastapi

from ..requesthandler import RequestHandler
from .iresource import IResource


class ResourceOptionsRequestHandler(RequestHandler[IResource]): # type: ignore
    allow: set[str]
    include_in_schema: bool = False

    def __init__(
        self,
        name: str,
        allow: set[str],
        include_in_schema: bool | None = True
    ):
        self.allow = allow
        super().__init__(
            name=name,
            method='OPTIONS',
            func=self.handle
        )

    async def handle(
        self,
        endpoint: IResource,
        *args: Any,
        **kwargs: Any
    ) -> fastapi.Response:
        return fastapi.Response(
            status_code=200,
            headers={
                'Allow': str.join(',', self.allow)
            }
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import datetime
from datetime import timezone
from typing import Any

import pydantic

from cbra.types import IResourceState


now = lambda: datetime.now(timezone.utc)


class ResourceMetadata(pydantic.BaseModel, IResourceState[Any]):
    created: datetime = pydantic.Field(default_factory=now)
    updated: datetime | None = pydantic.Field(default=None)
    deleted: datetime | None = pydantic.Field(default=None)
    version: str = pydantic.Field(default='')
    generation: int = pydantic.Field(default=0)


# Copyright (C) 2020-2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`GoogleTransport`."""
from typing import Any

from aorta.types import Envelope
from aorta.ext.google import PubsubTransport


class GooglePubsubTransport(PubsubTransport):
    prefix: str
    service_name: str

    def __init__(
        self,
        project: str,
        prefix: str,
        service_name: str
    ):
        super().__init__(
            project=project,
            topic=self.topic_factory,
            retry_topic=f'{self.prefix}.retry.{service_name}'
        )
        self.prefix = prefix
        self.service_name = service_name

    def topic_factory(self, envelope: Envelope[Any]) -> list[str]:
        topics: list[str] = [
            f'{self.prefix}.events',
            f'{self.prefix}.events.{envelope.kind}'
        ]
        if envelope.is_command():
            topics = [f'{self.prefix}.commands.{self.service_name}']
        return topics


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

import cbra.core as cbra
from .endpoint import AuthorizationServerEndpoint
from .types import AuthorizationRequest


class AuthorizationEndpoint(AuthorizationServerEndpoint):
    __module__: str = 'cbra.ext.oauth2'
    name: str = 'oauth2.authorize'
    status_code: int = 303
    summary: str = 'Authorization Endpoint'
    tags: list[str] = ['OAuth 2.x/OpenID Connect']

    async def get(
        self,
        params: AuthorizationRequest = AuthorizationRequest.depends()
    ) -> fastapi.Response:
        raise NotImplementedError
    
    @cbra.describe(summary="Authorization Endpoint (OpenID Connect)")
    async def post(self) -> None:
        """The OpenID Connect Core specification mandates that the **Authorization
        Endpoint** must support the HTTP `POST` method. This endpoint takes the
        parameters supported by the `GET` endpoint as the request body, which
        must be provided as `application/json` or `application/x-www-form-urlencoded`.

        *This endpoint is not implemented.*
        """
        # https://openid.net/specs/openid-connect-core-1_0.html#AuthRequest
        raise NotImplementedError


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import Generator


class PathParameter:
    __module__: str = 'cbra.types'
    format: str | None = None
    type: str

    @classmethod
    def __modify_schema__(
        cls,
        field_schema: dict[str, Any]
    ) -> None:
        field_schema.update( # pragma: no cover
            type=cls.type,
            format=cls.format
        )

    @classmethod
    def __get_validators__(cls) -> Generator[Callable[..., str | None], None, None]:
        yield cls.validate

    @classmethod
    def clean(cls, v: Any) -> Any:
        return v

    @classmethod
    def validate(cls, v: Any) -> str:
        # Called by FastAPI, must simply return the value.
        return v


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import TypeVar

import pydantic

from .resourceidentifier import ResourceIdentifier
from .resourcemodeltype import ResourceModelType


T = TypeVar('T', bound='ResourceModel')


class ResourceModel(pydantic.BaseModel, metaclass=ResourceModelType):
    __abstract__: bool = True
    __create_model__: type[pydantic.BaseModel]
    __update_model__: type[pydantic.BaseModel]
    __replace_model__: type[pydantic.BaseModel]
    __response_model__: type[pydantic.BaseModel]
    __list_model__: type[pydantic.BaseModel]
    __key_model__: type[ResourceIdentifier]

    def can_replace(self) -> bool:
        raise NotImplementedError

    def merge(self: T, resource: T) -> T:
        return self.parse_obj({
            **self.dict(),
            **resource.dict(
                exclude_defaults=True,
                exclude_none=True,
                exclude_unset=True
            ),
        })


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import copy
import inspect
import logging.config
from typing import Any
from typing import Callable

import aorta
import fastapi
from fastapi import FastAPI
from fastapi.routing import DecoratedCallable
from fastapi.params import Depends

from cbra.types import Abortable
from .conf import settings
from .endpoint import Endpoint
from .resource import Resource
from .ioc import Container
from .ioc import Requirement
from .messagepublisher import MessagePublisher
from .utils import parent_signature


class Application(FastAPI):
    __module__: str = 'cbra'
    _injectables: tuple[type, ...] = (
        Depends,
        Requirement
    )

    @parent_signature(FastAPI.__init__)
    def __init__(self, *args: Any, **kwargs: Any):
        self.container = Container.fromsettings()

        handlers: dict[type, Any] = kwargs.setdefault('exception_handlers', {})
        handlers[Abortable] = self.on_aborted

        kwargs.setdefault('root_path', settings.ASGI_ROOT_PATH)
        self.inject('MessagePublisher', MessagePublisher)
        self.inject('MessageTransport', aorta.NullTransport())
        super().__init__(*args, **kwargs)
        self.add_event_handler('startup', self.setup_logging) # type: ignore

    def add(
        self,
        routable: type[Endpoint | Resource],
        *args: Any, **kwargs: Any
    ) -> None:
        routable.add_to_router(self, *args, **kwargs)

    def inject(self, name: str, value: Any) -> None:
        """Inject a value into the dependencies container."""
        self.container.inject(name, value)

    def logging_config(self):
        config = copy.deepcopy(settings.LOGGING)
        if not settings.DEBUG and not settings.LOG_CONSOLE:
            # Remove console handler when not running in debug mode and its
            # not explicitely enabled in the settings.
            config['handlers']['console'] = {'class': 'logging.NullHandler'}
        return config

    def setup_logging(self) -> None:
        logging.config.dictConfig(self.logging_config())

    async def on_aborted(
        self,
        request: fastapi.Request,
        exc: Abortable
    ) -> fastapi.Response:
        return await exc.as_response()

    @parent_signature(FastAPI.add_api_route)
    def add_api_route(
        self,
        endpoint: Callable[..., Any],
        *args: Any,
        **kwargs: Any
    ) -> None:
        self.update_requirements(endpoint)
        return super().add_api_route(endpoint=endpoint, *args, **kwargs)

    @parent_signature(FastAPI.head)
    def head(self, *a: Any, **k: Any):
        return self.discover_requirements(FastAPI.head, *a, **k)

    @parent_signature(FastAPI.get)
    def get(self, *a: Any, **k: Any):
        return self.discover_requirements(FastAPI.get, *a, **k)

    @parent_signature(FastAPI.post)
    def post(self, *a: Any, **k: Any):
        return self.discover_requirements(FastAPI.post, *a, **k)

    @parent_signature(FastAPI.patch)
    def patch(self, *a: Any, **k: Any):
        return self.discover_requirements(FastAPI.patch, *a, **k)

    @parent_signature(FastAPI.put)
    def put(self, *a: Any, **k: Any):
        return self.discover_requirements(FastAPI.put, *a, **k)

    @parent_signature(FastAPI.trace)
    def trace(self, *a: Any, **k: Any):
        return self.discover_requirements(FastAPI.trace, *a, **k)

    @parent_signature(FastAPI.options)
    def options(self, *a: Any, **k: Any):
        return self.discover_requirements(FastAPI.options, *a, **k)

    @parent_signature(FastAPI.delete)
    def delete(self, *a: Any, **k: Any):
        return self.discover_requirements(FastAPI.delete, *a, **k)

    def discover_requirements(
        self,
        decorator_factory: Callable[..., DecoratedCallable],
        *args: Any,
        **kwargs: Any
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        def decorator(func: DecoratedCallable) -> DecoratedCallable:
            self.update_requirements(func)
            return decorator_factory(self, *args, **kwargs)(func)
        return decorator

    def update_requirements(self, func: Callable[..., Any] | Depends | Any) -> None:
        """Traverse the signature tree of the given function to find
        all :class:`Requirement` instances.
        """
        # TODO: this will completely mess up if multiple Application instances
        # are spawned.
        if not callable(func): return None
        if isinstance(func, Depends):
            return self.update_requirements(func.dependency)
        if isinstance(func, Requirement):
            func.add_to_container(self.container)
        signature = inspect.signature(func) # type: ignore
        for param in signature.parameters.values():
            if not isinstance(param.default, self._injectables):
                continue
            if isinstance(param.default, Requirement):
                param.default.add_to_container(self.container)
                if param.default.callable():
                    self.update_requirements(param.default.factory) # type: ignore
                continue

            if isinstance(param.default, Depends):
                injectable = param.default.dependency
                if injectable is None:
                    # Was declared as f(dependency: Callable = fastapi.Depends())
                    injectable = param.annotation
                self.update_requirements(injectable)


# Copyright (C) 2020-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import cast
from typing import Any

import aorta

from .eventarcendpoint import EventarcEndpoint
from .messagediscarded import MessageDiscarded
from .messagerunner import MessageRunner
from .pubsubmessage import PubsubMessage


class AortaEndpoint(EventarcEndpoint):
    __module__: str = 'cbra.ext.google'
    runner: MessageRunner = MessageRunner.depends()

    async def on_message(self, message: PubsubMessage) -> None:
        try:
            data = message.get_data()
        except ValueError:
            self.logger.critical("Data could not be interpreted as JSON.")
            raise MessageDiscarded
        if data is None:
            raise MessageDiscarded
        envelope = aorta.parse(data)
        if envelope is None:
            self.logger.critical("Message is not an Aorta message type.")
            raise MessageDiscarded
        if not envelope.is_known():
            self.logger.critical(
                "Received an Aorta message of unknown type "
                "(kind: %s, version: %s, id: %s)",
                envelope.kind, envelope.api_version, envelope.metadata.uid
            )
            raise MessageDiscarded
        assert isinstance(envelope, aorta.types.Envelope)
        await self.runner.run(cast(aorta.types.Envelope[Any], envelope))


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable

from .container import Container
from .dependency import Dependency
from .inheriteddependencydecorator import InheritDependenciesDecorator
from .injected import Injected
from .instance import Instance # type: ignore
from .requirement import Requirement
from .setting_ import Setting


__all__: list[str] = [
    'override',
    'inject',
    'instance',
    'setting',
    'Container',
    'Dependency',
    'Requirement',
]


def override(*args: Any, **kwargs: Any) -> Callable[..., Any]:
    """Injects the dependencies of the given callable."""
    return InheritDependenciesDecorator(*args, **kwargs)


def inject(name: str) -> Any:
    return Injected(name)


def instance(name: str, missing: Any = None) -> Any:
    return Instance(name, missing=missing)


def setting(name: str, default: Any = NotImplemented) -> Any:
    return Setting(name, missing=default)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import TypeVar
from typing import Union

import fastapi
import pydantic
from headless.ext import oauth2

from cbra.core.ioc import override
from cbra.types import IDependant
from .jarmauthorizeresponse import JARMAuthorizeResponse
from .queryauthorizeresponse import QueryAuthorizeResponse
from .responsevalidationfailure import ResponseValidationFailure
from .unsupportedauthorizationresponse import UnsupportedAuthorizationResponse


__all__: list[str] = ['RedirectParameters']
T = TypeVar('T', bound='_RedirectParameters')


class _RedirectParameters(IDependant, pydantic.BaseModel):

    @classmethod
    def fromrequest(
        cls: type[T],
        code: str | None = fastapi.Query(
            default=None,
            title='Authorization code',
            description=(
                'The authorization code generated by the authorization server. '
                'Required if using the Authorization Code Flow and the response '
                'mode is `query`, otherwise this parameter is ignored.'
            )
        ),
        state: str | None = fastapi.Query(
            default=None,
            title='State',
            description=(
                'An opaque value used by the client to maintain state '
                'between the request and callback. The authorization '
                'server includes this value when redirecting the '
                'user-agent back to the client. If supplied, this '
                'parameter **must** equal the `state` parameter used when '
                'creating the authorization request. The `state` parameter '
                'is ignored when using JARM because it is included in the '
                'JSON Web Token supplied using the `jwt` parameter per '
                'chosen response mode.'
            )
        ),
        iss: str | None = fastapi.Query(
            default=None,
            title='Issuer',
            description=(
                'Identifies the authorization server that redirected to '
                'this endpoint, as defined in RFC 9207. If the client supports '
                'OAuth 2.0 Authorization Server Issuer Identification, then '
                'this parameter is **required**, if the response mode is not '
                'JWT Secured Authorization Response Mode (JARM) per RFC 9101.\n\n'
                'The `iss` parameter value is the issuer identifier of '
                'the authorization server that created the authorization '
                'response, as defined in RFC 8414.  Its value **must** '
                'be a URL that uses the `https` scheme without any '
                'query or fragment components.'
            )
        )
    ) -> T:
        raise NotImplementedError

    @classmethod
    def __inject__(cls: type[T]) -> Callable[..., Awaitable[T] | T]:
        return cls.fromrequest


class RedirectParameters(_RedirectParameters):
    __root__: Union[
        QueryAuthorizeResponse,
        JARMAuthorizeResponse,
        UnsupportedAuthorizationResponse
    ]

    @classmethod
    @override(_RedirectParameters.fromrequest) # type: ignore
    def fromrequest(cls: type[T], **kwargs: Any) -> T: # type: ignore
        return cls.parse_obj(kwargs)

    async def obtain(
        self,
        client: oauth2.Client,
        **kwargs: Any
    ) -> oauth2.TokenResponse:
        return await self.__root__.obtain(client, **kwargs)
    
    def result(self) -> QueryAuthorizeResponse | JARMAuthorizeResponse:
        if isinstance(self.__root__, UnsupportedAuthorizationResponse):
            raise ResponseValidationFailure("Invalid response parameters.")
        return self.__root__


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .create import Create
from .collection import Collection
from .delete import Delete
from .queryresult import QueryResult
from .replace import Replace
from .resource import Resource
from .resourcemodel import ResourceModel
from .resourceprotocol import ResourceProtocol
from .resourcetype import ResourceType
from .retrieve import Retrieve
from .update import Update


__all__: list[str] = [
    'Create',
    'Collection',
    'Delete',
    'Mutable',
    'QueryResult',
    'Replace',
    'Resource',
    'ResourceModel',
    'ResourceProtocol',
    'ResourceType',
    'Retrieve',
    'Update',
]


class Mutable(
    Create,
    Delete,
    Replace,
    Retrieve,
    Update,
):
    __module__: str = 'cbra.core'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os

from cbra.types import Forbidden
from cbra.types import NullRequestPrincipal
from cbra.core import Endpoint

from .googleserviceaccountprincipal import GoogleServiceAccountPrincipal


class GoogleEndpoint(Endpoint):
    """A :class:`~cbra.core.Endpoint` implementation that is invoked
    by Google, for example by services such as Eventarc with Cloud Run,
    Cloud Tasks or Pubsub.
    """
    __abstract__: bool = True
    __module__: str = 'cbra.ext.google'

    #: The set of service account emails that may invoke this endpoint.
    allowed_service_accounts: set[str] = set()

    principal: GoogleServiceAccountPrincipal | NullRequestPrincipal # type: ignore
    require_authentication: bool = True
    trusted_providers: set[str] = {
        "https://accounts.google.com"
    }

    def __init__(self):
        super().__init__()
        if os.getenv('GOOGLE_SERVICE_ACCOUNT_EMAIL'):
            self.allowed_service_accounts.add(os.environ['GOOGLE_SERVICE_ACCOUNT_EMAIL'])

    async def authenticate(self) -> None:
        await super().authenticate()
        if not self.ctx.is_authenticated():
            return
        if self.ctx.subject.email not in self.allowed_service_accounts:
            self.logger.critical(
                'Google service endpoint was invoked with a disallowed '
                'service account (email: %s)',
                self.ctx.subject.email
            )
            raise Forbidden


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import fastapi


class IRoutable:
    __module__: str = 'cbra.types'
    response_model_by_alias: bool = False

    @classmethod
    def add_to_router(cls, endpoint: Any, router: fastapi.FastAPI, **kwargs: Any) -> None:
        raise NotImplementedError